from __future__ import annotations
import datetime
from typing import List
import pandas as pd
import os
import abc
from vdf_io.meta_types import NamespaceMeta, VDFMeta

from vdf_io.util import extract_data_hash, get_author_name, standardize_metric
from vdf_io.constants import ID_COLUMN


class ExportVDB(abc.ABC):
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if not hasattr(cls, "DB_NAME_SLUG"):
            raise TypeError(
                f"Class {cls.__name__} lacks required class variable 'DB_NAME_SLUG'"
            )

    def __init__(self, args):
        self.file_structure = []
        self.file_ctr = 1
        self.hash_value = extract_data_hash(args)
        self.args = args
        self.args["hash_value"] = self.hash_value
        self.args["exported_count"] = 0
        self.timestamp_in_format = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        self.vdf_directory = f"vdf_{self.timestamp_in_format}_{self.hash_value}"
        os.makedirs(self.vdf_directory, exist_ok=True)

    @abc.abstractmethod
    def get_index_names(self) -> List[str]:
        """
        Get index names from vector database
        """
        # raise NotImplementedError()
        pass

    @abc.abstractmethod
    def get_data(self) -> ExportVDB:
        """
        Get data from vector database
        """
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def make_parser(cls, subparsers):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def export_vdb(cls, args):
        raise NotImplementedError()

    def save_vectors_to_parquet(self, vectors, metadata, vectors_directory):
        vectors_df = pd.DataFrame(list(vectors.items()), columns=[ID_COLUMN, "vector"])

        if metadata:
            metadata_list = [{**{ID_COLUMN: k}, **v} for k, v in metadata.items()]
            metadata_df = pd.DataFrame.from_records(metadata_list)

            # Check for duplicate column names and rename as necessary
            common_columns = set(vectors_df.columns) & set(metadata_df.columns) - {
                ID_COLUMN
            }
            metadata_df.rename(
                columns={col: f"metadata_{col}" for col in common_columns}, inplace=True
            )

            df = vectors_df.merge(metadata_df, on=ID_COLUMN, how="outer")
        else:
            df = vectors_df

        parquet_file = os.path.join(vectors_directory, f"{self.file_ctr}.parquet")
        df.to_parquet(parquet_file)
        self.file_structure.append(parquet_file)
        self.file_ctr += 1

        vectors = {}
        metadata = {}
        return len(df)

    def get_basic_vdf_meta(self, index_metas):
        return VDFMeta(
            version=self.args["library_version"],
            file_structure=self.file_structure,
            author=get_author_name(),
            exported_from=self.DB_NAME_SLUG,
            indexes=index_metas,
            exported_at=datetime.datetime.now().astimezone().isoformat(),
        )

    def get_namespace_meta(
        self,
        index_name,
        vectors_directory,
        total,
        num_vectors_exported,
        dim,
        index_config=None,
        vector_columns=None,
        distance=None,
    ):
        namespace_meta = NamespaceMeta(
            index_name=index_name,
            namespace="",
            total_vector_count=total,
            exported_vector_count=num_vectors_exported,
            metric=standardize_metric(
                distance,
                self.DB_NAME_SLUG,
            ),
            dimensions=dim,
            model_name=self.args["model_name"],
            vector_columns=["vector"] if vector_columns is None else vector_columns,
            data_path="/".join(vectors_directory.split("/")[1:]),
            index_config=index_config,
        )

        return namespace_meta



#!/usr/bin/env python3

import argparse
import datetime
import json
import os
import time
import litellm
from litellm import EmbeddingResponse
import numpy as np
import sentence_transformers
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_random_exponential,
)
import torch
from tqdm import tqdm
from dotenv import load_dotenv
import sys
from IPython.core import ultratb
import warnings
import pyarrow as pa
import pyarrow.parquet as pq
from mlx_embedding_models.embedding import EmbeddingModel
from sentence_transformers import SentenceTransformer

import vdf_io
from vdf_io.constants import ID_COLUMN
from vdf_io.meta_types import NamespaceMeta, VDFMeta

from vdf_io.util import (
    get_author_name,
    get_final_data_path,
    get_parquet_files,
    read_parquet_progress,
    set_arg_from_input,
    set_arg_from_password,
)

litellm.suppress_debug_info = True
warnings.filterwarnings("ignore", module="litellm")
sys.excepthook = ultratb.FormattedTB(color_scheme="Linux", call_pdb=False)

load_dotenv()


def main():
    start_time = time.time()
    reembed()
    end_time = time.time()
    print(f"Time taken: {end_time - start_time:.2f} seconds")


def reembed():
    parser = argparse.ArgumentParser(description="Reembed a vector dataset")

    add_arguments_to_parser(parser)

    args = parser.parse_args()
    args = vars(args)

    if args.get("env_file_path"):
        load_dotenv(os.path.join(os.getcwd(), args.get("env_file_path")), override=True)

    take_input_from_cli_prompt(args)

    reembed_count = 0
    # open VDF_META.json
    vdf_meta_path = os.path.join(args["dir"], "VDF_META.json")
    if not os.path.exists(vdf_meta_path) or not valid_json(vdf_meta_path):
        handle_new_dataset(args)
    reembed_impl(args, reembed_count)


def valid_json(file_path):
    try:
        with open(file_path) as f:
            json.load(f)
            return True
    except Exception as e:
        print(f"Error reading {file_path}: {e}")
        return False


def reembed_impl(args, reembed_count):
    with open(os.path.join(args["dir"], "VDF_META.json"), "r+") as f:
        vdf_meta = json.load(f)
        for _, index_meta in tqdm(
            vdf_meta["indexes"].items(), desc="Iterating over indexes"
        ):
            overwrite_bool = args["overwrite"]
            for namespace_meta in tqdm(index_meta, desc="Iterating over namespaces"):
                data_path = namespace_meta["data_path"]
                final_data_path = get_final_data_path(
                    os.getcwd(), args["dir"], data_path, args={}
                )
                parquet_files = get_parquet_files(
                    final_data_path,
                    args={},
                    id_column=vdf_meta.get("id_column", ID_COLUMN),
                )
                for file in tqdm(parquet_files, desc="Iterating over parquet files"):
                    file_path = os.path.join(final_data_path, file)
                    if "vector_columns" not in namespace_meta:
                        namespace_meta["vector_columns"] = []
                    df = read_parquet_progress(file_path, ID_COLUMN)
                    if args["text_column"] not in df.columns:
                        ask_for_text_column(args, file_path, df)
                    new_vector_column = (
                        f"vec_{args['text_column']}_{args['new_model_name'].replace('/', '_')}"
                        + (
                            "_" + args["quantize"]
                            if (args.get("quantize") != "float32")
                            else ""
                        )
                    )
                    if "dimensions" in args and args["dimensions"] is not None:
                        new_vector_column += f"_dim{args['dimensions']}"
                    else:
                        litellm_embed_response = call_litellm(args, ["dummy"])
                        dims = len(litellm_embed_response.data[0]["embedding"])
                        new_vector_column += f"_dim{dims}"
                    if (
                        new_vector_column in namespace_meta["vector_columns"]
                        and not overwrite_bool
                    ):
                        # ask user if they want to overwrite (y/n)
                        overwrite = args["overwrite"] or input(
                            f"{new_vector_column} already exists in vector_columns. Overwrite? (y/n): "
                        )
                        overwrite_bool = overwrite.lower() == "y"
                        if not overwrite_bool:
                            tqdm.write(
                                f"Skipping {file_path} because {new_vector_column} already exists in vector_columns.\n"
                                "Aborting reembedding."
                            )
                            exit()
                    tqdm.write(f"Reembedding {file_path}")
                    # read parquet file
                    # get text column
                    # get embeddings
                    BATCH_SIZE = args.get("batch_size")
                    all_embeddings = []
                    pbar = tqdm(total=len(df), desc="Iterating over rows")
                    for i in tqdm(
                        range(0, len(df), BATCH_SIZE), desc="Iterating over batches"
                    ):
                        batch_text = df[args["text_column"]][
                            i : i + BATCH_SIZE
                        ].tolist()
                        pbar.update(len(batch_text))

                        embeddings = call_litellm(args, batch_text)
                        # add embeddings to df
                        # embeddings.data is a list of dicts. Each dict has keys "embedding" and "index"
                        # first sort by "index" and then extract "embedding"
                        vectors = [
                            x["embedding"]
                            for x in sorted(embeddings.data, key=lambda x: x["index"])
                        ]
                        dim = len(vectors[0])
                        all_embeddings.extend(vectors)
                    # add only the new_vector_column in parquet file
                    df[new_vector_column] = all_embeddings
                    if ID_COLUMN not in df.columns:
                        df[ID_COLUMN] = df.index
                    df.to_parquet(file_path)
                    tqdm.write(
                        f"Computed {len(all_embeddings)} vectors for {len(df)} rows in column:{new_vector_column} of {file_path}"
                    )
                    reembed_count += len(all_embeddings)
                # prepend new_vector_column to vector_columns in namespace_meta
                if new_vector_column not in namespace_meta["vector_columns"]:
                    namespace_meta["vector_columns"].insert(0, new_vector_column)
                else:
                    tqdm.write(
                        f"Warning: {new_vector_column} already exists in vector_columns. Overwriting."
                    )
                if not namespace_meta.get("model_map"):
                    namespace_meta["model_map"] = {}
                for vector_column in namespace_meta["vector_columns"]:
                    if vector_column not in namespace_meta["model_map"]:
                        namespace_meta["model_map"][vector_column] = {
                            "model_name": namespace_meta.get("model_name"),
                            "text_column": args["text_column"],
                            "dimensions": namespace_meta.get("dimensions"),
                            "vector_column": vector_column,
                        }
                namespace_meta["model_map"][new_vector_column] = {
                    "model_name": args["new_model_name"],
                    "text_column": args["text_column"],
                    "dimensions": dim,
                    "vector_column": new_vector_column,
                }
                namespace_meta["model_name"] = args["new_model_name"]
                namespace_meta["dimensions"] = dim
                namespace_meta["schema_dict_str"] = pq.read_schema(
                    file_path
                ).to_string()
        # write vdf_meta to VDF_META.json
        f.seek(0)
        json.dump(vdf_meta, f, indent=4)
        tqdm.write(
            f"Reembedding complete. Computed {reembed_count} vectors. Updated VDF_META.json"
        )


def ask_for_text_column(args, file_path, df):
    text_columns = args["text_column"].split("|")
    if all([col in df.columns for col in text_columns]):
        df[args["text_column"]] = df[text_columns].apply(
            lambda x: " ".join(x.dropna().astype(str)), axis=1
        )
    else:
        # ask user to enter text columns by index1|index2|index3
        tqdm.write(f"Text column {args['text_column']} not found in {file_path}")
        tqdm.write("Select the text column(s) from the following list:")
        # list of string columns
        text_column_options = {}
        for i, col in enumerate(df.columns):
            # check if column is of type string
            # pick first non-null value
            non_null_value = df[col].dropna().iloc[0]
            if isinstance(non_null_value, str):
                tqdm.write(f"{i+1}: {col}")
                text_column_options[i + 1] = col
        choice_correctly_entered = False
        while not choice_correctly_entered:
            text_column_choice = input(
                "Enter the text column index(es) separated by '|' (e.g. 1|2|3): "
            )
            text_columns_choice = [
                int(col_no) for col_no in text_column_choice.split("|")
            ]
            if all([tc in text_column_options for tc in text_columns_choice]):
                choice_correctly_entered = True
                text_column_choice_names = [
                    text_column_options[col_no] for col_no in text_columns_choice
                ]
                text_column_name = "|".join(text_column_choice_names)
                tqdm.write(f"Selected text column(s): {text_column_name}.")
                args["text_column"] = text_column_name
                if len(text_column_choice_names) > 1:
                    tqdm.write("Combining multiple text columns into a single column.")
                    df[text_column_name] = df[text_column_choice_names].apply(
                        lambda x: " ".join(x.dropna().astype(str)),
                        axis=1,
                    )
            else:
                tqdm.write(
                    "Invalid choice. Please enter the text column index(es) separated by '|': "
                )


def handle_new_dataset(args):
    # create VDF_META.json
    with open(os.path.join(args["dir"], "VDF_META.json"), "w") as f:
        # index_name is folder name of dir
        index_name = os.path.basename(args["dir"])
        # find parquet files in dir recursively
        import glob

        parquet_files = glob.glob(
            os.path.join(args["dir"], "**", "*.parquet"), recursive=True
        )
        if len(parquet_files) == 0:
            raise Exception("No parquet files found in the specified directory")
        total_vector_count = 0
        for file in parquet_files:
            # read schema of parquet file using pyarrow and get count of rows
            import pyarrow.parquet as pq

            table = pq.read_table(file)
            row_count = table.num_rows
            total_vector_count += row_count
            # print each column name and type
            # for col in schema_dict["metadata"]["arrow:extension:column_types"]:
            #     tqdm.write(f"Column: {col['name']} of type {col['type']}")
            # # are there any vector columns in the schema?
            # for col in schema_dict["metadata"]["arrow:extension:vector_columns"]:
            #     tqdm.write(f"Vector column: {col['name']} of type {col['type']}")
            # create VDF_META.json
            # if there exists a vector column, add it to vector_columns
            vector_columns = []
            model_map = {}
            model_name = ""
            dimensions = -1
            for field in table.schema:
                if pa.types.is_list(field.type):
                    tqdm.write(
                        f"Vector column: {field.name} of type {field.type.to_pandas_dtype()}"
                    )
                    vector_columns.append(field.name)
                    # check length of vector column
                    dimensions = len(table[field.name].to_pandas().iloc[0])
                    # ask which model to use for embedding
                    tqdm.write(
                        f"Vector column {field.name} has length {dimensions}. Please specify the model to use for embedding."
                    )
                    model_name = input(
                        "Enter the name of the model to be used for embedding: "
                    )
                    old_text_column = input(
                        "Enter the name of the column containing text to be embedded: "
                    )
                    model_map[field.name] = {
                        "model_name": model_name,
                        "text_column": old_text_column,
                        "dimensions": dimensions,
                        "vector_column": field.name,
                    }
                    # add model_name to VDF_META.json
                    # add vector column to vector_columns
                    # add vector column to model_map
                    # add model_name to model_map
                    # add dimensions to model_map

        vdf_meta = VDFMeta(
            version=vdf_io.__version__,
            file_structure=[],
            author=get_author_name(),
            exported_from="reembed",
            exported_at=datetime.datetime.now().astimezone().isoformat(),
            id_column=ID_COLUMN,
            indexes={
                index_name: [
                    NamespaceMeta(
                        namespace="",
                        index_name=index_name,
                        total_vector_count=total_vector_count,
                        exported_vector_count=total_vector_count,
                        dimensions=dimensions,
                        model_name=model_name,
                        vector_columns=vector_columns,
                        data_path=".",
                        metric=None,
                        model_map=model_map,
                        schema_dict_str=table.schema.to_string(),
                    )
                ]
            },
        )
        vdf_meta = vdf_meta.model_dump()
        json.dump(vdf_meta, f, indent=4)


def take_input_from_cli_prompt(args):
    set_arg_from_input(
        args,
        "new_model_name",
        "Enter the name of the new model to be used (default: 'text-embedding-3-small'): ",
        str,
        "text-embedding-3-small",
    )

    if args["new_model_name"].startswith("text-embedding-3"):
        set_arg_from_input(
            args,
            "dimensions",
            "Enter the dimensions of the new model to be used (default: 1536 for small, 3072 for large): ",
            int,
        )

    set_arg_from_password(
        args,
        "OPENAI_API_KEY",
        "Enter the OpenAI API key (hit return to skip): ",
        "OPENAI_API_KEY",
    )

    set_arg_from_input(
        args,
        "text_column",
        "Enter the name of the column containing text to be embedded (default: 'text'): ",
        str,
        "text",
    )

    set_arg_from_input(
        args,
        "dir",
        "Enter the directory of vector dataset in the VDF format: ",
        str,
    )


def add_arguments_to_parser(parser):
    parser.add_argument(
        "-d",
        "--dir",
        type=str,
        help="Directory of vector dataset in the VDF format",
        required=True,
    )

    parser.add_argument(
        "-m",
        "--new_model_name",
        type=str,
        help="Name of new model to be used",
    )

    parser.add_argument(
        "-t",
        "--text_column",
        type=str,
        help="Name of the column containing text to be embedded",
        default="text",
    )

    parser.add_argument(
        "--dimensions",
        type=int,
        help="Dimensions of the new model to be used",
    )

    parser.add_argument(
        "--env_file_path",
        type=str,
        help="Path to the .env file",
    )

    parser.add_argument(
        "--batch_size",
        type=int,
        help="Batch size for reembedding",
        default=96,
    )

    parser.add_argument(
        "--overwrite",
        type=bool,
        help="Overwrite existing vector columns",
        default=False,
        action=argparse.BooleanOptionalAction,
    )

    parser.add_argument(
        "--input_type",
        type=str,
        help="Input type for the model",
        default=None,
    )
    parser.add_argument(
        "--disable_mlx",
        type=bool,
        help="Disable MLX for embedding",
        default=False,
        action=argparse.BooleanOptionalAction,
    )

    quantize_options = [
        "int8",
        "binary",
        "float32",
        "int8",
        "uint8",
        "binary",
        "ubinary",
    ]
    parser.add_argument(
        "--quantize",
        type=str,
        choices=quantize_options,
        help=f"Quantization method ({', '.join(quantize_options)})",
        default="float32",
    )


use_sentence_transformers = False


@retry(
    wait=wait_random_exponential(multiplier=1, max=10),
    stop=stop_after_attempt(5),
    retry=retry_if_exception_type(Exception),
)
def call_litellm(args, batch_text):
    if args["new_model_name"].startswith("huggingface/"):
        return call_sentence_transformers(args, batch_text)
    global use_sentence_transformers
    if use_sentence_transformers:
        return call_sentence_transformers(args, batch_text)
    try:
        embeddings = litellm.embedding(
            **{
                "model": args["new_model_name"],
                "input": batch_text,
                "dimensions": args.get("dimensions"),
                **(
                    {"input_type": args["input_type"]}
                    if args.get("input_type") is not None
                    else {}
                ),
            }
        )
        if args.get("quantize") != "float32":
            # convert to ndarray
            embeddings = np.array([np.array(x["embedding"]) for x in embeddings.data])
            embeddings = sentence_transformers.quantize_embeddings(
                embeddings, precision=args.get("quantize")
            )
            # convert to EmbeddingResponse
            embeddings = EmbeddingResponse(
                data=[
                    {"index": i, "embedding": emb.tolist()}
                    for i, emb in enumerate(embeddings)
                ]
            )
    except Exception as e:
        # catch BadRequestError: LLM Provider NOT provided. Pass in the LLM provider you are trying to call. You passed model=TaylorAI/bge-micro-v2
        # check type of exception
        if e.message.startswith("Huggingface") or (
            type(e).__name__ == "BadRequestError" and "LLM Provider" in e.message
        ):
            tqdm.write(
                f"Using Sentence Transformers for embedding for model {args['new_model_name']} {e}"
            )
            use_sentence_transformers = True
            embeddings = call_sentence_transformers(args, batch_text)
        else:
            raise e

    return embeddings


model = None

using_mlx = False


def is_apple_silicon():
    # if `uname -m` returns arm64, then it is an apple silicon device
    return os.uname().machine == "arm64"
    # return torch.backends.mps.is_available()


def call_sentence_transformers(args, batch_text):
    global model
    if model is None:
        # figure out device map
        device = "cuda" if torch.cuda.is_available() else "cpu"
        if device == "cpu" and is_apple_silicon() and not args.get("disable_mlx"):
            global using_mlx
            using_mlx = True
            tqdm.write("Using MLX for embedding")
            model = EmbeddingModel.from_registry(strip_hf_prefix(args))
        else:
            model = SentenceTransformer(
                strip_hf_prefix(args),
                trust_remote_code=True,
                device=device,
            )
    if using_mlx:
        embeddings = model.encode(batch_text)
    else:
        embeddings = model.encode(batch_text, precision=args.get("quantize"))
    return EmbeddingResponse(
        data=[
            {"index": i, "embedding": emb.tolist()} for i, emb in enumerate(embeddings)
        ]
    )


def strip_hf_prefix(args):
    # count /s
    if args["new_model_name"].count("/") == 2:
        # replace first occurrence of "huggingface/" to ""
        return args["new_model_name"].replace("huggingface/", "", 1)
    return args["new_model_name"].replace("huggingface/", "")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

from getpass import getpass
import os
from huggingface_hub import HfApi
import argparse


def push_to_hub(export_obj, args):
    args = vars(args)
    print("Pushing to HuggingFace Hub...")

    # Log in to Hugging Face
    if (
        "HUGGING_FACE_TOKEN" not in os.environ
        or os.environ["HUGGING_FACE_TOKEN"] is None
    ):
        # set HUGGINGFACEHUB_API_TOKEN env var
        os.environ["HUGGING_FACE_TOKEN"] = getpass(
            prompt="Enter your HuggingFace API token (with write access): "
        )
    if (
        "HF_USERNAME" not in args
        or args["HF_USERNAME"] is None
        or args["HF_USERNAME"] == ""
    ):
        if "HF_USERNAME" not in os.environ or os.environ["HF_USERNAME"] is None:
            # set HF_USERNAME env var
            os.environ["HF_USERNAME"] = input("Enter your HuggingFace username: ")
    else:
        os.environ["HF_USERNAME"] = args["HF_USERNAME"]
    hf_api = HfApi(token=os.environ["HUGGING_FACE_TOKEN"])
    data_path = os.path.join(os.getcwd(), export_obj.vdf_directory)
    export_obj.vdf_directory = os.path.basename(export_obj.vdf_directory.rstrip("/"))

    if args["name"] is not None:
        args["name"] = "vdf_" + args["name"]
    else:
        args["name"] = export_obj.vdf_directory.replace("/", "_")
    repo_id = f"{os.environ['HF_USERNAME']}/{args['name']}"
    print(f"Creating a new dataset at {repo_id}")
    dataset_url = hf_api.create_repo(
        token=os.environ["HUGGING_FACE_TOKEN"],
        repo_id=repo_id,
        private=(not args["public"]),
        repo_type="dataset",
        exist_ok=True,
    )
    # for each file/folder in export_obj.vdf_directory, upload to hub
    hf_api.upload_folder(
        repo_id=repo_id,
        folder_path=data_path,
        repo_type="dataset",
    )
    # create hf dataset card in temp README.md
    readme_path = os.path.join(data_path, "README.md")
    with open(readme_path, "w") as f:
        f.write(
            """
---
tags:
- vdf
- vector-io
- vector-dataset
- vector-embeddings
---
This is a dataset created using [vector-io](https://github.com/ai-northstar-tech/vector-io)
"""
        )
    hf_api.upload_file(
        repo_id=repo_id,
        path_or_fileobj=readme_path,
        path_in_repo="README.md",
        repo_type="dataset",
    )
    print(
        f"Created a {'public' if args['public'] else 'private'} dataset at {dataset_url}"
    )


def main():
    parser = argparse.ArgumentParser(
        description="Push a vector dataset to HuggingFace Hub"
    )
    parser.add_argument(
        "-d",
        "--vdf_directory",
        type=str,
        help="Path to the directory containing the vector dataset",
    )
    parser.add_argument(
        "--name",
        type=str,
        help="Name of the dataset (default: name of the directory)",
        default=None,
    )
    parser.add_argument(
        "--public",
        type=bool,
        help="Make the dataset public (default: False)",
        default=False,
        action=argparse.BooleanOptionalAction,
    )
    parser.add_argument(
        "--HF_USERNAME",
        type=str,
        help="HuggingFace username (default: existing env var HF_USERNAME)",
        default=None,
    )
    args = parser.parse_args()

    if args.vdf_directory is None:
        print("Please provide a path to the vector dataset directory")
        args.vdf_directory = input("Enter the path to the vector dataset directory: ")

    push_to_hub(args, args)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import pyarrow.parquet as pq
import os
import argparse


def get_file_size_in_gb(file_path):
    """
    Calculate the size of a file in gigabytes.

    Args:
        file_path (str): The path to the file.

    Returns:
        float: The size of the file in gigabytes.
    """
    return os.path.getsize(file_path) / (1024 * 1024 * 1024)  # Convert bytes to GB


def main():
    """
    Consolidate multiple Parquet files into combined files based on a maximum size.

    Example command:
    python count_rows.py /path/to/parquet_files
    """
    # Directory containing your parquet files
    parser = argparse.ArgumentParser(description="Consolidate Parquet files")
    parser.add_argument(
        "--directory",
        type=str,
        help="Directory containing your parquet files",
        default=None,
    )

    args = parser.parse_args()
    args = vars(args)
    if "directory" not in args or args["directory"] is None:
        args["directory"] = input("Enter the directory containing your parquet files: ")
        args["directory"] = os.path.join(os.getcwd(), args["directory"])
    directory = args["directory"]
    cnt = 0
    for filename in os.listdir(directory):
        if filename.endswith(".parquet"):
            filepath = os.path.join(directory, filename)
            table = pq.read_table(filepath)
            cnt += table.num_rows
    print("row count", cnt)


if __name__ == "__main__":
    main()



class DBNames:
    PINECONE = "pinecone"
    QDRANT = "qdrant"
    MILVUS = "milvus"
    VERTEXAI = "vertexai_vectorsearch"
    KDBAI = "kdbai"
    VESPA = "vespa"
    LANCEDB = "lancedb"
    TXTAI = "txtai"
    CHROMA = "chroma"
    WEAVIATE = "weaviate"
    ASTRADB = "astradb"



"""
Import data to vertex ai vector search index
"""

import argparse
from typing import Dict, List
import uuid
import time
import numpy as np
import json
import itertools
from tqdm import tqdm
from ratelimit import limits, sleep_and_retry
from backoff import on_exception, expo

# gcloud config set project $PROJECT_ID - users
# SCOPES = ["https://www.googleapis.com/auth/cloud-platform"]

# NEW
from google.cloud import aiplatform as aip
import google.cloud.aiplatform_v1 as aipv1
from google.cloud import storage
from google.protobuf import struct_pb2
from google.cloud.aiplatform_v1.types.index import Index
from google.cloud.aiplatform_v1.types.index_endpoint import IndexEndpoint
from google.cloud.aiplatform_v1.types.index_endpoint import DeployedIndex
import google.api_core.exceptions as google_exceptions

from vdf_io.names import DBNames
from vdf_io.import_vdf.vdf_import_cls import ImportVDB
from vdf_io.util import read_parquet_progress, set_arg_from_input
from vdf_io.constants import ID_COLUMN, INT_MAX


# exceptions
class ResourceNotExistException(Exception):
    def __init__(self, resource: str, message="Resource Does Not Exist."):
        self.resource = resource
        self.message = message
        super().__init__(self.message)


class ImportVertexAIVectorSearch(ImportVDB):
    DB_NAME_SLUG = DBNames.VERTEXAI

    @classmethod
    def import_vdb(cls, args):
        """
        Import data to Vertex AI Vector Search
        """
        set_arg_from_input(args, "project_id", "Enter the Google Cloud Project ID:")
        set_arg_from_input(args, "location", "Enter the region hosting your index: ")
        set_arg_from_input(
            args,
            "requests_per_minute",
            "Optional. Enter desired requests per minute for rate limit (default: 6000): ",
            default_value=6000,
            type_name=int,
        )
        set_arg_from_input(
            args,
            "filter_restricts",
            "Optional. Enter list of dicts describing string filters for each data point: ",
        )
        set_arg_from_input(
            args,
            "numeric_restricts",
            "Optional. Enter list of dicts for each datapoint: ",
        )
        set_arg_from_input(
            args,
            "crowding_tag",
            "Optional. CrowdingTag of the datapoint: ",
            type_name=str,
        )
        if args["create_new"] is True:
            set_arg_from_input(
                args,
                "gcs_bucket",
                "Optional. Enter valid gcs bucket name (or one will be created per index_name): ",
                type_name=str,
            )
            set_arg_from_input(
                args,
                "approx_nn_count",
                "Optional. The default number of neighbors to find via approximate search (default: 150): ",
                type_name=int,
                default_value=150,
            )
            set_arg_from_input(
                args,
                "leaf_node_emb_count",
                "Optional. Number of embeddings on each leaf node (default: 1000): ",
                type_name=int,
                default_value=1000,
            )
            set_arg_from_input(
                args,
                "leaf_nodes_percent",
                "Optional. The default percentage of leaf nodes that any query may be searched (default: 10 [10%]): ",
                type_name=int,
                default_value=10,
                # choices=range(1, 101)
            )
            set_arg_from_input(
                args,
                "distance_measure",
                "Optional. The distance measure used in nearest neighbor search (default: `DOT_PRODUCT_DISTANCE`): ",
                type_name=str,
                default_value="DOT_PRODUCT_DISTANCE",
                # choices=[
                #     "DOT_PRODUCT_DISTANCE",
                #     "COSINE_DISTANCE",
                #     "L1_DISTANCE",
                #     "SQUARED_L2_DISTANCE"
                # ],
            )
            set_arg_from_input(
                args,
                "shard_size",
                "Optional. Size of the shards (default: `SHARD_SIZE_MEDIUM`): ",
                type_name=str,
                default_value="SHARD_SIZE_MEDIUM",
                # choices=[
                #     "SHARD_SIZE_SMALL",
                #     "SHARD_SIZE_MEDIUM",
                #     "SHARD_SIZE_LARGE",
                # ],
            )
        if args["deploy_new_index"] is True:
            set_arg_from_input(
                args,
                "machine_type",
                "Optional. The type of machine (default: `e2-standard-16`): ",
                type_name=str,
                default_value="e2-standard-16",
                # choices=[
                #     "n1-standard-16",
                #     "n1-standard-32",
                #     "e2-standard-2",
                #     "e2-standard-16",
                #     "e2-highmem-16",
                #     "n2d-standard-32",
                # ],
            )
            set_arg_from_input(
                args,
                "min_replicas",
                "Optional. The minimum number of machine replicas for deployed index (default: 1): ",
                type_name=int,
                default_value=1,
            )
            set_arg_from_input(
                args,
                "max_replicas",
                "Optional. The maximum number of machine replicas for deployed index (default: 1): ",
                type_name=int,
                default_value=1,
            )

        vertexai_vectorsearch_import = ImportVertexAIVectorSearch(args)
        vertexai_vectorsearch_import.upsert_data()
        return vertexai_vectorsearch_import

    @classmethod
    def make_parser(cls, subparsers):
        parser_vertexai_vectorsearch = subparsers.add_parser(
            DBNames.VERTEXAI, help="Import data to Vertex AI Vector Search"
        )
        parser_vertexai_vectorsearch.add_argument(
            "-p", "--project-id", type=str, help="Google Cloud Project ID"
        )
        parser_vertexai_vectorsearch.add_argument(
            "-l", "--location", type=str, help="Google Cloud region hosting your index"
        )
        parser_vertexai_vectorsearch.add_argument(
            "-f", "--filter-restricts", type=str, help="string filters"
        )
        parser_vertexai_vectorsearch.add_argument(
            "-n", "--numeric-restricts", type=str, help="numeric filters"
        )
        parser_vertexai_vectorsearch.add_argument(
            "-r", "--requests-per-minute", type=int, help="rate limiter"
        )
        parser_vertexai_vectorsearch.add_argument(
            "-c",
            "--crowding-tag",
            type=str,
            help="string value to enforce diversity in retrieval",
        )
        parser_vertexai_vectorsearch.add_argument(
            "--deploy_new_index",
            type=bool,
            help="deploy new index (default: False)",
            default=False,
            action=argparse.BooleanOptionalAction,
        )

    def __init__(self, args: Dict) -> None:
        super().__init__(args)
        self.DB_NAME_SLUG = DBNames.VERTEXAI
        self.project_id = self.args["project_id"]
        self.location = self.args["location"]
        self.batch_size = self.args.get("batch_size", 100) or 100

        # optional
        self.create_new_index = self.args.get("create_new_index", False)

        self.deploy_new_index = self.args.get("deploy_new_index", False)

        # =========================================================
        # set index, endpoint, and SDK client
        # =========================================================
        self.parent = f"projects/{self.project_id}/locations/{self.location}"

        client_endpoint = f"{self.location}-aiplatform.googleapis.com"
        # set index client
        self.index_client = aipv1.IndexServiceClient(
            client_options=dict(api_endpoint=client_endpoint)
        )
        # set index endpoint client
        self.index_endpoint_client = aipv1.IndexEndpointServiceClient(
            client_options=dict(api_endpoint=client_endpoint)
        )
        # set SDK client
        aip.init(
            project=self.project_id,
            location=self.location,
        )
        # cloud storage client
        self.storage_client = storage.Client(project=self.project_id)

        # =========================================================
        # filters: restricts and crowding
        # =========================================================
        if self.args["filter_restricts"]:
            # String filters: allows and denies
            allows = []
            denies = []
            list_restrict_entries = []
            for name in self.args["filter_restricts"]:
                name_space_filter_entry = {}
                all_allows = []
                all_denies = []
                allows = []
                denies = []
                name_space_filter_entry["namespace"] = name.get("namespace")
                if name.get("allow_list") is not None:
                    allow_items = name.get("allow_list")
                    allows.append(allow_items)
                    # allows.append([a for a in allow_items])
                if name.get("deny_list") is not None:
                    deny_items = name.get("deny_list")
                    denies.append(deny_items)

                if allows:
                    all_allows = list(itertools.chain.from_iterable(allows))
                    name_space_filter_entry["allow_list"] = all_allows

                if denies:
                    all_denies = list(itertools.chain.from_iterable(denies))
                    name_space_filter_entry["deny_list"] = all_denies

                list_restrict_entries.append(name_space_filter_entry)

        self.list_restrict_entries = (
            list_restrict_entries if self.args["filter_restricts"] is not None else None
        )
        print(f"list_restrict_entries : {self.list_restrict_entries}")

        if self.args["numeric_restricts"]:
            # Numeric filters:
            list_of_numeric_entries = []
            for name in self.args["numeric_restricts"]:
                name_space_filter_entry = {}
                name_space_filter_entry["namespace"] = name.get("namespace")
                name_space_filter_entry["data_type"] = name.get("data_type")
                list_of_numeric_entries.append(name_space_filter_entry)

        self.list_of_numeric_entries = (
            list_of_numeric_entries
            if self.args["numeric_restricts"] is not None
            else None
        )
        print(f"list_of_numeric_entries : {self.list_of_numeric_entries}")

        # =========================================================
        # set index args
        # =========================================================
        self.index_names = []
        for index_name, index_meta in self.vdf_meta["indexes"].items():
            # load data
            print(f"Importing data for index: {index_name}")
            print(f"index_meta: {json.dumps(index_meta, indent=4)}")
            for namespace_meta in index_meta:
                self.set_dims(namespace_meta, index_name)

                all_indexes = [index.display_name for index in self.list_indexes()]
                # check if index exists
                index_name = index_name + (
                    f"_{namespace_meta['namespace']}"
                    if namespace_meta["namespace"]
                    else ""
                )
                suffix = 2
                while index_name in all_indexes and self.args["create_new"] is True:
                    index_name = index_name + f"_{suffix}"
                    suffix += 1
                self.index_name = index_name
                self.index_names.append(self.index_name)
                if self.index_name not in all_indexes:
                    # create index
                    unique_id = uuid.uuid4()
                    if self.index_name is None:
                        self.index_name = f"my_vvs_index_{unique_id}"
                        print(f"Creating new index: {self.index_name} ...")
                    print(f"dimensions: {namespace_meta['dimensions']}")

                    # optional; used if create_new_index = True
                    self.approx_nn_count = self.args.get("approx_nn_count", 150)
                    self.leaf_node_emb_count = self.args.get(
                        "leaf_node_emb_count", 1000
                    )
                    self.leaf_nodes_percent = self.args.get("leaf_nodes_percent", 7)
                    self.distance_measure = self.args.get(
                        "distance_measure", "DOT_PRODUCT_DISTANCE"
                    )
                    self.shard_size = self.args.get("shard_size", "SHARD_SIZE_MEDIUM")
                    self.gcs_bucket = self.args.get("gcs_bucket", None)

                    if self.gcs_bucket is None:
                        self.gcs_bucket = self.index_name
                        try:
                            print(f"Creating new gcs_bucket: {self.gcs_bucket} ...")
                            bucket_client = self.storage_client.bucket(self.gcs_bucket)
                            new_bucket = self.storage_client.create_bucket(
                                self.gcs_bucket, location=self.location
                            )
                            print(
                                f"Created bucket {new_bucket.name} in {new_bucket.location}"
                            )
                        except Exception as e:
                            print(f"{self.gcs_bucket} bucket already exists {e}")
                            pass
                    self.gcs_folder = "init_index"
                    self.local_file_name = "embeddings_0.json"
                    self.contents_delta_uri = (
                        f"gs://{self.gcs_bucket}/{self.gcs_folder}"
                    )

                    # dummy embedding - TODO: use input data from parquet(?)
                    init_embedding = {
                        ID_COLUMN: str(unique_id),
                        "embedding": list(np.zeros(namespace_meta["dimensions"])),
                    }

                    # dump embedding to a local file
                    with open(self.local_file_name, "w") as f:
                        json.dump(init_embedding, f)

                    # upload to GCS
                    bucket_client = self.storage_client.bucket(self.gcs_bucket)
                    blob = bucket_client.blob(
                        f"{self.gcs_folder}/{self.local_file_name}"
                    )
                    blob.upload_from_filename(f"{self.local_file_name}")

                    self.target_index = self._create_index(
                        index_name=self.index_name,
                        contents_delta_uri=self.contents_delta_uri,
                        dimensions=namespace_meta["dimensions"],
                        approximate_neighbors_count=self.approx_nn_count,
                        leaf_node_embedding_count=self.leaf_node_emb_count,
                        leaf_nodes_to_search_percent=self.leaf_nodes_percent,
                        distance_measure_type=self.distance_measure,
                        shard_size=self.shard_size,
                    )

                    if self.deploy_new_index:
                        # Note: Don't need to deploy index to upsert vectors to index

                        self.index_endpoint_name = f"{self.index_name}_endpoint"
                        self.index_endpoint = self._create_index_endpoint(
                            endpoint_name=self.index_endpoint_name
                        )

                        # optional; used if deploy_new_index = True
                        self.machine_type = self.args.get(
                            "machine_type", "e2-standard-16"
                        )
                        self.min_replicas = self.args.get("min_replicas", 1)
                        self.max_replicas = self.args.get("max_replicas", 1)

                        self._deploy_index(
                            index_name=self.index_name,
                            endpoint_name=self.index_endpoint.display_name,
                            machine_type=self.machine_type,
                            min_replicas=self.min_replicas,
                            max_replicas=self.max_replicas,
                        )
                else:
                    self.target_index = self._get_index()
                    if self.target_index is None:
                        raise ValueError(
                            f"Failed to set target_index: {self.index_name}"
                        )

                self.target_index_resource_name = self.target_index.name

                # init target index to import vectors to
                self.target_vertexai_index = aip.MatchingEngineIndex(
                    self.target_index_resource_name
                )
                print(f"Importing to index: {self.target_vertexai_index.display_name}")
                print(f"Full resource name: {self.target_vertexai_index.resource_name}")
                print("Target index config:")

                index_config_dict = self.target_vertexai_index.to_dict()
                _index_meta_config = index_config_dict["metadata"]["config"]
                tqdm.write(json.dumps(_index_meta_config, indent=4))

    # =========================================================
    # VectorSearch helpers
    # =========================================================
    def _set_index_name(self, index_name: str) -> None:
        """

        :param index_name:
        :return:
        """
        self.index_name = index_name

    def _set_index_endpoint_name(self, index_endpoint_name: str = None) -> None:
        """

        :param index_endpoint_name:
        :return:
        """
        if index_endpoint_name is not None:
            self.index_endpoint_name = index_endpoint_name
        elif self.index_name is not None:
            self.index_endpoint_name = f"{self.index_name}_endpoint"
        else:
            raise ResourceNotExistException("index")

    def _get_index(self) -> Index:
        """

        :return:
        """
        # Check if index exists
        d_ids = []
        indexes = []
        if self.index_name is not None:
            all_indexes = [index for index in self.list_indexes()]
            try:
                indexes = [
                    index.name
                    for index in all_indexes
                    if index.display_name == self.index_name
                    or index.name == self.index_name
                ]
            except Exception as e:
                print(
                    f"{self.index_name} not an existing display_name or resource_name: {e}"
                )
                pass
            if not indexes:
                try:
                    # checking deployed indexes
                    for test_index in all_indexes:
                        if test_index.deployed_indexes:
                            # grabbing all deployed indexes
                            d_ids.extend(test_index.deployed_indexes)

                    # returning only those that match
                    deployed_index_match = [
                        d_id
                        for d_id in d_ids
                        if (
                            d_id.display_name == self.index_name
                            or d_id.deployed_index_id == self.index_name
                        )
                    ]
                    # need to do this to return the index resoruce_name
                    if deployed_index_match:
                        target_endpoint = aip.MatchingEngineIndexEndpoint(
                            deployed_index_match[0].index_endpoint
                        )
                        for d in target_endpoint.deployed_indexes:
                            if (
                                d.id == self.index_name
                                or d.display_name == self.index_name
                            ):
                                indexes.append(d.index)
                except Exception as e:
                    print(f"not an existing deployed_index: {e}")
                    pass

        if len(indexes) == 0:
            print(f"Index {self.index_name} not found")
            return None
        else:
            index_id = indexes[0]
            # print("Found existing index")
            request = aipv1.GetIndexRequest(name=index_id)
            index = self.index_client.get_index(request=request)
            return index

    def _get_index_endpoint(self) -> IndexEndpoint:
        """

        :return:
        """
        # Check if index endpoint exists
        all_index_endpoints = [response for response in self.list_index_endpoints()]

        if self.index_endpoint_name is not None:
            print(f"Checking if {self.index_endpoint_name} already exists...")
            try:
                index_endpoints = [
                    response.name
                    for response in all_index_endpoints
                    if response.display_name == self.index_endpoint_name
                    or response.name == self.index_endpoint_name
                ]
            except Exception as e:
                print(f"{self.index_endpoint_name} not an existing index endpoint: {e}")
                pass
        else:
            raise ResourceNotExistException("index_endpoint")

        if len(index_endpoints) == 0:
            print(f"Could not find index endpoint: {self.index_endpoint_name}")
            return None
        else:
            index_endpoint_id = index_endpoints[0]
            index_endpoint = self.index_endpoint_client.get_index_endpoint(
                name=index_endpoint_id
            )
            return index_endpoint

    def list_indexes(self) -> List[Index]:
        """

        :return:
        """
        request = aipv1.ListIndexesRequest(parent=self.parent)
        page_result = self.index_client.list_indexes(request=request)
        indexes = [response for response in page_result]
        return indexes

    def list_index_endpoints(self) -> List[IndexEndpoint]:
        """

        :return:
        """
        request = aipv1.ListIndexEndpointsRequest(parent=self.parent)
        page_result = self.index_endpoint_client.list_index_endpoints(request=request)
        index_endpoints = [response for response in page_result]
        return index_endpoints

    def list_deployed_indexes(self, endpoint_name: str = None) -> List[DeployedIndex]:
        """

        :param endpoint_name:
        :return:
        """
        try:
            if endpoint_name is not None:
                self._set_index_endpoint_name(index_endpoint_name=endpoint_name)
            index_endpoint = self._get_index_endpoint()
            deployed_indexes = index_endpoint.deployed_indexes
        except ResourceNotExistException as rnee:
            raise rnee

        return list(deployed_indexes)

    def _build_index_config(
        self,
        contents_delta_uri: str,
        index_display_name: str,
        dimensions: int,
        approximate_neighbors_count: int,
        leaf_node_embedding_count: int,
        leaf_nodes_to_search_percent: int,
        description: str,
        distance_measure_type: str,
        shard_size: str,
    ):
        vector_search_index_config = {
            "index_display_name": index_display_name,
            "contents_delta_uri": contents_delta_uri,
            "dimensions": dimensions,
            "approximate_neighbors_count": approximate_neighbors_count,
            "distance_measure_type": distance_measure_type,
            "leaf_node_embedding_count": leaf_node_embedding_count,
            "leaf_nodes_to_search_percent": leaf_nodes_to_search_percent,
            "description": description,
            "labels": {
                "project": f"{self.project_id}",
                "tag": "vectorio-import",
            },
        }
        print("vector_search_index_config:")
        tqdm.write(json.dumps(vector_search_index_config, indent=4))

        tree_ah_config = struct_pb2.Struct(
            fields={
                "leafNodeEmbeddingCount": struct_pb2.Value(
                    number_value=vector_search_index_config["leaf_node_embedding_count"]
                ),
                "leafNodesToSearchPercent": struct_pb2.Value(
                    number_value=vector_search_index_config[
                        "leaf_nodes_to_search_percent"
                    ]
                ),
            }
        )
        algorithm_config = struct_pb2.Struct(
            fields={"treeAhConfig": struct_pb2.Value(struct_value=tree_ah_config)}
        )
        config = struct_pb2.Struct(
            fields={
                "dimensions": struct_pb2.Value(
                    number_value=vector_search_index_config["dimensions"]
                ),
                "approximateNeighborsCount": struct_pb2.Value(
                    number_value=vector_search_index_config[
                        "approximate_neighbors_count"
                    ]
                ),
                "distanceMeasureType": struct_pb2.Value(
                    string_value=vector_search_index_config["distance_measure_type"]
                ),
                "algorithmConfig": struct_pb2.Value(struct_value=algorithm_config),
                "shardSize": struct_pb2.Value(string_value=shard_size),
            }
        )
        metadata = struct_pb2.Struct(
            fields={
                "config": struct_pb2.Value(struct_value=config),
                "contentsDeltaUri": struct_pb2.Value(
                    string_value=vector_search_index_config["contents_delta_uri"]
                ),
            }
        )
        return metadata

    def _create_index(
        self,
        contents_delta_uri: str,
        dimensions: int,
        approximate_neighbors_count: int,
        leaf_node_embedding_count: int,
        leaf_nodes_to_search_percent: int,
        distance_measure_type: str,
        shard_size: str,
        index_name: str = None,
    ) -> Index:
        """

        :param contents_delta_uri:
        :param dimensions:
        :param approximate_neighbors_count:
        :param leaf_node_embedding_count:
        :param leaf_nodes_to_search_percent:
        :param distance_measure_type:
        :param shard_size:
        :param index_name:
        :return:
        """

        if index_name is not None:
            self._set_index_name(index_name=index_name)
        # Get index
        if self.index_name is None:
            raise ResourceNotExistException("index")
        index = self._get_index()
        # Create index if does not exists
        if index:
            print(f"Index {self.index_name} exists with resource_name:\n {index.name}")
        else:
            print(f"Index {self.index_name} does not exists. Creating index ...")

            invoke_time = time.strftime("%Y%m%d_%H%M%S")
            description = f"created during vectorio import at {invoke_time}"

            metadata = self._build_index_config(
                contents_delta_uri=contents_delta_uri,
                index_display_name=index_name,
                dimensions=dimensions,
                description=description,
                approximate_neighbors_count=approximate_neighbors_count,
                leaf_node_embedding_count=leaf_node_embedding_count,
                leaf_nodes_to_search_percent=leaf_nodes_to_search_percent,
                distance_measure_type=distance_measure_type,
                shard_size=shard_size,
            )

            index_request = {
                "display_name": self.index_name,
                "description": description,
                "metadata": struct_pb2.Value(struct_value=metadata),
                "index_update_method": aipv1.Index.IndexUpdateMethod.STREAM_UPDATE,
            }

            r = self.index_client.create_index(
                parent=self.parent, index=Index(index_request)
            )

            # Poll the operation until its done successfully.
            print("Poll the operation to create index ...")
            while True:
                if r.done():
                    break
                time.sleep(5)
                print(".", end="")

            index = r.result()
            print(
                f"\nIndex {self.index_name} created with resource_name:\n {index.name}"
            )

        return index

    def _create_index_endpoint(
        self,
        endpoint_name: str = None,
    ) -> IndexEndpoint:
        """

        :param endpoint_name:
        :return:
        """
        try:
            if endpoint_name is not None:
                self._set_index_endpoint_name(index_endpoint_name=endpoint_name)

            # Get index endpoint if exists
            index_endpoint = self._get_index_endpoint()

            # Create Index Endpoint if does not exists
            if index_endpoint is not None:
                print("Index endpoint already exists")
            else:
                print(
                    f"Index endpoint {self.index_endpoint_name} does not exists. Creating index endpoint..."
                )

                index_endpoint_request = {"display_name": self.index_endpoint_name}

                index_endpoint = IndexEndpoint(index_endpoint_request)
                index_endpoint.public_endpoint_enabled = True

                r = self.index_endpoint_client.create_index_endpoint(
                    parent=self.parent, index_endpoint=index_endpoint
                )

                print("Poll the operation to create index endpoint ...")
                while True:
                    if r.done():
                        break
                    time.sleep(10)
                    print(".", end="")

                index_endpoint = r.result()

        except Exception as e:
            print(f"Failed to create index endpoint {self.index_endpoint_name}")
            raise e

        print(f"\nReturning index endpoint: {index_endpoint}")
        return index_endpoint

    def _deploy_index(
        self,
        index_name: str = None,
        endpoint_name: str = None,
        machine_type: str = "e2-standard-16",
        min_replicas: int = 2,
        max_replicas: int = 2,
    ) -> IndexEndpoint:
        """

        :param endpoint_name:
        :param index_name:
        :param machine_type:
        :param min_replicas:
        :param max_replicas:
        :return:
        """

        index = self._get_index()
        index_endpoint = self._get_index_endpoint()

        # Deploy Index to endpoint
        try:
            # Check if index is already deployed to the endpoint
            if index_endpoint.deployed_indexes:
                if index.name in index_endpoint.deployed_indexes:
                    print(
                        f"Skipping Index deployment. Index {self.index_name}"
                        + f"already deployed {index.name} to endpoint {self.index_endpoint_name}"
                    )
                    return index_endpoint

            invoke_time = time.strftime("%Y%m%d_%H%M%S")
            deployed_index_id = f"{self.index_name.replace('-', '_')}_{invoke_time}"
            deploy_index_config = {
                ID_COLUMN: deployed_index_id,
                "display_name": deployed_index_id,
                "index": index.name,
                "dedicated_resources": {
                    "machine_spec": {
                        "machine_type": machine_type,
                    },
                    "min_replica_count": min_replicas,
                    "max_replica_count": max_replicas,
                },
            }
            print("Deploying index with request:")
            tqdm.write(json.dumps(deploy_index_config, indent=4))
            r = self.index_endpoint_client.deploy_index(
                index_endpoint=index_endpoint.name,
                deployed_index=DeployedIndex(deploy_index_config),
            )

            # Poll the operation until its done successfullly.
            print("Poll the operation to deploy index ...")
            while True:
                if r.done():
                    break
                time.sleep(60)
                print(".", end="")

            print(
                f"\nDeployed {self.index_name} to endpoint {self.index_endpoint_name}"
            )

        except Exception as e:
            print(
                f"Failed to deploy {self.index_name} to endpoint: {self.index_endpoint_name}"
            )
            raise e

        return index_endpoint

    def upsert_data(self):
        max_hit = False
        self.total_imported_count = 0
        MINUTE = 60
        CALLS_PER_PRD = self.args.get("requests_per_minute", 6000)

        # rate is 1 QPS
        @sleep_and_retry  # If there are more requests than rate, sleep shortly
        @on_exception(
            expo, google_exceptions.ResourceExhausted, max_tries=10
        )  # if we receive exceptions from Google API, retry
        @limits(calls=CALLS_PER_PRD, period=MINUTE)
        def upsert_in_rate(self, upsert_request):
            """
            Upserts vectors and applies rate limits.
            """
            return self.index_client.upsert_datapoints(request=upsert_request)

        # upsert for each index
        for new_index_name, (index_name, index_meta) in tqdm(
            zip(self.index_names, self.vdf_meta["indexes"].items()),
            desc="Iterating over indexes",
        ):
            self.index_name = new_index_name
            self.target_index = self._get_index()
            self.target_index_resource_name = self.target_index.name
            # init target index to import vectors to
            self.target_vertexai_index = aip.MatchingEngineIndex(
                self.target_index_resource_name
            )
            # load data
            tqdm.write(
                f"Importing data from {index_name} to {self.target_vertexai_index.display_name}"
            )
            tqdm.write(f"index_meta: {json.dumps(index_meta, indent=4)}")

            for namespace_meta in tqdm(index_meta, desc="Iterating over namespaces"):
                # get data path
                data_path = namespace_meta["data_path"]
                final_data_path = self.get_final_data_path(data_path)
                print(f"data_path: {data_path}")

                # get col names
                vector_metadata_names, vector_column_name = self.get_vector_column_name(
                    namespace_meta["vector_columns"], namespace_meta
                )
                print(f"vector_column_name    : {vector_column_name}")
                print(f"vector_metadata_names : {vector_metadata_names}")

                # Load the data from the parquet files
                parquet_files = self.get_parquet_files(final_data_path)

                total_ids = []
                for file in tqdm(parquet_files, desc="Iterating over parquet files"):
                    file_path = self.get_file_path(final_data_path, file)
                    df = read_parquet_progress(file_path, self.id_column)
                    df[ID_COLUMN] = df[ID_COLUMN].apply(lambda x: str(x))

                    insert_datapoints_payload = []

                    for idx, row in tqdm(
                        df.iterrows(), desc="Iterating over rows", total=len(df)
                    ):
                        row = json.loads(row.to_json())

                        total_ids.append(row[ID_COLUMN])
                        row[vector_column_name] = self.extract_vector(
                            row[vector_column_name]
                        )
                        numeric_restrict_entry_list = []
                        restrict_entry_list = []
                        allow_values = []
                        deny_values = []
                        crowding_tag_val = None

                        # if idx == 10:
                        #     # sanity check
                        #     print(f"row['id'] : {row['id']}")

                        if self.list_restrict_entries:
                            for entry in self.list_restrict_entries:
                                restrict_entry = {}

                                restrict_entry["namespace"] = entry.get("namespace")

                                if entry.get("allow_list"):
                                    for col in entry.get("allow_list"):
                                        allow_values.append(row[col])
                                        restrict_entry["allow_list"] = [
                                            str(a) for a in allow_values
                                        ]

                                if entry.get("deny_list"):
                                    for col in entry.get("deny_list"):
                                        deny_values.append(row[col])
                                        restrict_entry["deny_list"] = [
                                            str(d) for d in deny_values
                                        ]

                                restrict_entry_list.append(restrict_entry)

                                # if idx == 10:
                                #     print(f"restrict_entry_list : {restrict_entry_list}")

                        if self.list_of_numeric_entries:
                            # numeric_restrict_entry_list = []
                            for entry in self.list_of_numeric_entries:
                                numeric_restrict_entry = {}

                                data_type = entry.get("data_type")
                                col_name = entry.get("namespace")
                                numeric_restrict_entry["namespace"] = entry.get(
                                    "namespace"
                                )
                                numeric_restrict_entry[data_type] = row[col_name]
                                numeric_restrict_entry_list.append(
                                    numeric_restrict_entry
                                )

                            # if idx == 10:
                            #     # sanity check
                            #     print(f"numeric_restrict_entry_list : {numeric_restrict_entry_list}")

                        if self.args["crowding_tag"]:
                            crowding_tag_col = self.args["crowding_tag"]
                            crowding_tag_val = str(row[crowding_tag_col])

                            # if idx == 10:
                            #     # sanity check
                            #     print(f"crowding_tag_col : {crowding_tag_col}")
                            #     print(f"crowding_tag_val : {crowding_tag_val}")

                        insert_datapoints_payload.append(
                            aipv1.IndexDatapoint(
                                datapoint_id=row[ID_COLUMN],
                                feature_vector=row[vector_column_name],
                                restricts=restrict_entry_list,
                                numeric_restricts=numeric_restrict_entry_list,
                                crowding_tag=aipv1.IndexDatapoint.CrowdingTag(
                                    crowding_attribute=crowding_tag_val
                                ),
                            )
                        )

                        if idx % self.batch_size == 0:
                            upsert_request = aipv1.UpsertDatapointsRequest(
                                index=self.target_vertexai_index.resource_name,
                                datapoints=insert_datapoints_payload,
                            )
                            if self.total_imported_count + len(
                                upsert_request.datapoints
                            ) >= (self.args.get("max_num_rows") or INT_MAX):
                                upsert_request = aipv1.UpsertDatapointsRequest(
                                    index=self.target_vertexai_index.resource_name,
                                    datapoints=insert_datapoints_payload[
                                        : (
                                            (self.args.get("max_num_rows") or INT_MAX)
                                            - self.total_imported_count
                                        )
                                    ],
                                )
                                max_hit = True
                            # self.index_client.upsert_datapoints(request=upsert_request)
                            upsert_in_rate(self, upsert_request=upsert_request)
                            self.total_imported_count += len(upsert_request.datapoints)
                            insert_datapoints_payload = []
                            if max_hit:
                                break
                        if max_hit:
                            break

                    if len(insert_datapoints_payload) > 0:
                        upsert_request = aipv1.UpsertDatapointsRequest(
                            index=self.target_vertexai_index.resource_name,
                            datapoints=insert_datapoints_payload,
                        )
                        if self.total_imported_count + len(
                            upsert_request.datapoints
                        ) >= (self.args.get("max_num_rows") or INT_MAX):
                            upsert_request = aipv1.UpsertDatapointsRequest(
                                index=self.target_vertexai_index.resource_name,
                                datapoints=insert_datapoints_payload[
                                    : (
                                        (self.args.get("max_num_rows") or INT_MAX)
                                        - self.total_imported_count
                                    )
                                ],
                            )
                            max_hit = True
                        # self.index_client.upsert_datapoints(request=upsert_request)
                        upsert_in_rate(self, upsert_request=upsert_request)
                        self.total_imported_count += len(upsert_request.datapoints)
                    if max_hit:
                        tqdm.write(
                            f"Max rows to be imported {self.args['max_num_rows']} hit. Exiting"
                        )
                        break
        print("Index import complete")
        print(
            f"Updated {self.target_vertexai_index.display_name} with {len(total_ids)} vectors"
        )
        self.args["imported_count"] = self.total_imported_count



import argparse
import json
from typing import Dict, List
from qdrant_client import QdrantClient
import os
from tqdm import tqdm
from dotenv import load_dotenv

from vdf_io.export_vdf.vdb_export_cls import ExportVDB
from vdf_io.names import DBNames
from vdf_io.meta_types import NamespaceMeta
from vdf_io.util import set_arg_from_input, set_arg_from_password

load_dotenv()

MAX_FETCH_SIZE = 1_000


class ExportQdrant(ExportVDB):
    DB_NAME_SLUG = DBNames.QDRANT

    @classmethod
    def make_parser(cls, subparsers):
        parser_qdrant = subparsers.add_parser(
            cls.DB_NAME_SLUG, help="Export data from Qdrant"
        )
        parser_qdrant.add_argument(
            "-u", "--url", type=str, help="Location of Qdrant instance"
        )
        parser_qdrant.add_argument(
            "-c", "--collections", type=str, help="Names of collections to export"
        )
        parser_qdrant.add_argument(
            "--prefer_grpc",
            type=bool,
            help="Whether to use GRPC. Recommended. (default: True)",
            default=True,
            action=argparse.BooleanOptionalAction,
        )

    @classmethod
    def export_vdb(cls, args):
        """
        Export data from Qdrant
        """
        set_arg_from_input(
            args,
            "url",
            "Enter the URL of Qdrant instance (default: 'http://localhost:6334'): ",
            str,
            "http://localhost:6334",
        )
        set_arg_from_input(
            args,
            "prefer_grpc",
            "Whether to use GRPC. Recommended. (default: True): ",
            bool,
            True,
        )
        set_arg_from_input(
            args,
            "collections",
            "Enter the name of collection(s) to export (comma-separated) (hit return to export all):",
            str,
        )
        set_arg_from_password(
            args, "qdrant_api_key", "Enter your Qdrant API key: ", "QDRANT_API_KEY"
        )
        qdrant_export = ExportQdrant(args)
        qdrant_export.get_data()
        return qdrant_export

    def __init__(self, args):
        """
        Initialize the class
        """
        super().__init__(args)
        self.client = QdrantClient(
            url=self.args["url"],
            api_key=self.args.get("qdrant_api_key", None),
            prefer_grpc=self.args.get("prefer_grpc", True),
        )

    def get_index_names(self) -> List[str]:
        """
        Get all collection names from Qdrant
        """
        collections = self.client.get_collections().collections
        collection_names = [collection.name for collection in collections]
        return collection_names

    def get_data(self):
        if "collections" not in self.args or self.args["collections"] is None:
            collection_names = self.get_index_names()
        else:
            collection_names = self.args["collections"].split(",")

        index_metas: Dict[str, List[NamespaceMeta]] = {}
        for collection_name in tqdm(collection_names, desc="Fetching indexes"):
            index_meta = self.get_data_for_collection(collection_name)
            index_metas[collection_name] = index_meta

        # Create and save internal metadata JSON
        self.file_structure.append(os.path.join(self.vdf_directory, "VDF_META.json"))
        internal_metadata = self.get_basic_vdf_meta(index_metas)
        meta_text = json.dumps(internal_metadata.model_dump(), indent=4)
        tqdm.write(meta_text)
        with open(os.path.join(self.vdf_directory, "VDF_META.json"), "w") as json_file:
            json_file.write(meta_text)
        # print internal metadata properly
        return True

    def try_scroll(self, fetch_size, collection_name, next_offset):
        try:
            records, next_offset = self.client.scroll(
                collection_name=collection_name,
                offset=next_offset,
                limit=fetch_size,
                with_payload=True,
                with_vectors=True,
                shard_key_selector=self.args.get("shard_key_selector", None),
            )
            return records, next_offset, fetch_size
        except Exception as e:
            # if it is keyboard interrupt, raise it
            if isinstance(e, KeyboardInterrupt):
                raise e
            tqdm.write(
                f"Failed to fetch data, reducing fetch size to{(fetch_size * 2) // 3}"
            )
            return self.try_scroll((fetch_size * 2) // 3, collection_name, next_offset)

    def get_data_for_collection(self, collection_name) -> List[NamespaceMeta]:
        vectors_directory = os.path.join(self.vdf_directory, collection_name)
        os.makedirs(vectors_directory, exist_ok=True)

        total = self.client.get_collection(collection_name).vectors_count

        num_vectors_exported = 0
        dim = self.client.get_collection(collection_name).config.params.vectors.size
        next_offset = 0
        records, next_offset, fetch_size = self.try_scroll(
            MAX_FETCH_SIZE, collection_name, next_offset
        )
        num_vectors_exported += self.save_from_records(
            records,
            vectors_directory,
        )
        pbar = tqdm(total=total, desc=f"Exporting {collection_name}")
        while next_offset is not None:
            records, next_offset, fetch_size = self.try_scroll(
                fetch_size, collection_name, next_offset
            )
            num_vectors_exported += self.save_from_records(
                records,
                vectors_directory,
            )
            pbar.update(len(records))

        namespace_meta = self.get_namespace_meta(
            collection_name,
            vectors_directory,
            total,
            num_vectors_exported,
            dim,
            index_config=self.client.get_collection(
                collection_name
            ).config.model_dump(),
            distance=self.client.get_collection(
                collection_name
            ).config.params.vectors.distance,
        )
        self.args["exported_count"] += num_vectors_exported
        return [namespace_meta]

    def save_from_records(self, records, vectors_directory):
        num_vectors_exported = 0
        vectors = {}
        metadata = {}
        for point in records:
            vectors[point.id] = point.vector
            metadata[point.id] = point.payload
        num_vectors_exported += self.save_vectors_to_parquet(
            vectors, metadata, vectors_directory
        )
        return num_vectors_exported



import json
import os
from typing import Dict, List
import lancedb
import pandas as pd
import pyarrow
from tqdm import tqdm
from vdf_io.meta_types import NamespaceMeta

from vdf_io.names import DBNames
from vdf_io.util import set_arg_from_input, set_arg_from_password
from vdf_io.export_vdf.vdb_export_cls import ExportVDB


class ExportLanceDB(ExportVDB):
    DB_NAME_SLUG = DBNames.LANCEDB

    @classmethod
    def make_parser(cls, subparsers):
        parser_lancedb = subparsers.add_parser(
            cls.DB_NAME_SLUG, help="Export data from LanceDB"
        )

        parser_lancedb.add_argument(
            "--endpoint", type=str, help="Location of LanceDB instance"
        )
        parser_lancedb.add_argument(
            "--lancedb_api_key", type=str, help="LanceDB API key"
        )
        parser_lancedb.add_argument(
            "--tables", type=str, help="LanceDB tables to export (comma-separated)"
        )
        parser_lancedb.add_argument(
            "--batch_size",
            type=int,
            help="Batch size for exporting data",
            default=10_000,
        )

    @classmethod
    def export_vdb(cls, args):
        set_arg_from_input(
            args,
            "endpoint",
            "Enter the URL of LanceDB instance (default: '~/.lancedb'): ",
            str,
            "~/.lancedb",
        )
        set_arg_from_password(
            args,
            "lancedb_api_key",
            "Enter the LanceDB API key: ",
            "LANCEDB_API_KEY",
        )
        set_arg_from_input(
            args,
            "tables",
            "Enter the name of tables to export (comma-separated, all will be exported by default): ",
            str,
            None,
        )
        lancedb_export = ExportLanceDB(args)
        lancedb_export.get_data()
        return lancedb_export

    def __init__(self, args):
        super().__init__(args)
        self.db = lancedb.connect(
            self.args["endpoint"], api_key=self.args.get("lancedb_api_key") or None
        )

    def get_index_names(self):
        if self.args.get("tables", None) is not None:
            return self.args["tables"].split(",")
        return self.db.table_names()

    def get_data(self):
        index_names = self.get_index_names()
        BATCH_SIZE = self.args["batch_size"]
        total = 0
        index_metas: Dict[str, List[NamespaceMeta]] = {}
        for index_name in index_names:
            namespace_metas = []
            vectors_directory = os.path.join(self.vdf_directory, index_name)
            os.makedirs(vectors_directory, exist_ok=True)
            table = self.db.open_table(index_name)
            offset = 0
            remainder_df = None
            j = 0
            for batch in tqdm(table.to_lance().to_batches()):
                df = batch.to_pandas()
                if remainder_df is not None:
                    df = pd.concat([remainder_df, df])
                while len(df) >= BATCH_SIZE:
                    df[:BATCH_SIZE].to_parquet(
                        os.path.join(vectors_directory, f"{j}.parquet")
                    )
                    j += 1
                    total += BATCH_SIZE
                    df = df[BATCH_SIZE:]
                offset += BATCH_SIZE
                remainder_df = df
            if remainder_df is not None and len(remainder_df) > 0:
                remainder_df.to_parquet(os.path.join(vectors_directory, f"{j}.parquet"))
                total += len(remainder_df)
            dim = -1
            for name in table.schema.names:
                if pyarrow.types.is_fixed_size_list(table.schema.field(name).type):
                    dim = table.schema.field(name).type.list_size
            vector_columns = [
                name
                for name in table.schema.names
                if pyarrow.types.is_fixed_size_list(table.schema.field(name).type)
            ]
            distance = "Cosine"
            try:
                for index in table.list_indices():
                    if index.vector_column_name == vector_columns[0]:
                        distance = vector_columns[0]
            except Exception:
                pass

            namespace_metas = [
                self.get_namespace_meta(
                    index_name,
                    vectors_directory,
                    total=total,
                    num_vectors_exported=total,
                    dim=dim,
                    vector_columns=vector_columns,
                    distance=distance,
                )
            ]
            index_metas[index_name] = namespace_metas
        self.file_structure.append(os.path.join(self.vdf_directory, "VDF_META.json"))
        internal_metadata = self.get_basic_vdf_meta(index_metas)
        meta_text = json.dumps(internal_metadata.model_dump(), indent=4)
        tqdm.write(meta_text)
        with open(os.path.join(self.vdf_directory, "VDF_META.json"), "w") as json_file:
            json_file.write(meta_text)
        # print internal metadata properly
        return True



import argparse
import pandas as pd
from tqdm import tqdm
import os
from dotenv import load_dotenv

from pinecone import Pinecone, ServerlessSpec, PodSpec, Vector

from vdf_io.constants import INT_MAX
from vdf_io.names import DBNames
from vdf_io.util import (
    set_arg_from_input,
    set_arg_from_password,
    standardize_metric_reverse,
)
from vdf_io.import_vdf.vdf_import_cls import ImportVDB

load_dotenv()


class ImportPinecone(ImportVDB):
    DB_NAME_SLUG = DBNames.PINECONE

    @classmethod
    def import_vdb(cls, args):
        """
        Import data to Pinecone
        """
        set_arg_from_password(
            args,
            "pinecone_api_key",
            "Enter your Pinecone API key: ",
            "PINECONE_API_KEY",
        )
        if args["serverless"] is False:
            set_arg_from_input(
                args, "environment", "Enter the environment of Pinecone instance: "
            )
        else:
            set_arg_from_input(
                args,
                "cloud",
                "Enter the cloud of Pinecone Serverless instance (default: 'aws'): ",
                str,
                "aws",
            )
            set_arg_from_input(
                args,
                "region",
                "Enter the region of Pinecone Serverless instance (default: 'us-west-2'): ",
                str,
                "us-west-2",
            )

        if args["subset"] is True:
            if "id_list_file" not in args or args["id_list_file"] is None:
                set_arg_from_input(
                    args,
                    "id_range_start",
                    "Enter the start of id range (hit return to skip): ",
                    int,
                )
                if args["id_range_start"] is not None:
                    set_arg_from_input(
                        args,
                        "id_range_end",
                        "Enter the end of id range (hit return to skip): ",
                        int,
                    )
            if args.get("id_range_start") is None and args.get("id_range_end") is None:
                set_arg_from_input(
                    args,
                    "id_list_file",
                    "Enter the path to id list file (hit return to skip): ",
                )

        pinecone_import = ImportPinecone(args)
        pinecone_import.upsert_data()
        return pinecone_import

    @classmethod
    def make_parser(cls, subparsers):
        parser_pinecone = subparsers.add_parser(
            cls.DB_NAME_SLUG, help="Import data to Pinecone"
        )
        parser_pinecone.add_argument(
            "-e", "--environment", type=str, help="Pinecone environment"
        )
        parser_pinecone.add_argument(
            "--serverless",
            type=bool,
            help="Import data to Pinecone Serverless (default: False)",
            default=False,
            action=argparse.BooleanOptionalAction,
        )
        parser_pinecone.add_argument(
            "-c", "--cloud", type=str, help="Pinecone Serverless cloud"
        )
        parser_pinecone.add_argument(
            "-r", "--region", type=str, help="Pinecone Serverless region"
        )

    def __init__(self, args):
        super().__init__(args)
        self.pc = Pinecone(api_key=self.args["pinecone_api_key"])

    def compliant_name(self, name: str) -> str:
        new_name = name.lower().replace("_", "-")
        return new_name

    def upsert_data(self):
        max_hit = False
        # Iterate over the indexes and import the data
        for index_name, index_meta in tqdm(
            self.vdf_meta["indexes"].items(), desc="Importing indexes"
        ):
            tqdm.write(f"Importing data for index '{index_name}'")
            for namespace_meta in index_meta:
                self.set_dims(namespace_meta, index_name)
            # list indexes
            indexes = self.pc.list_indexes().names()
            # check if index exists
            index_name = self.create_new_name(index_name, indexes)
            compliant_index_name = self.compliant_name(index_name)
            if compliant_index_name not in indexes:
                # create index
                try:
                    if self.args["serverless"] is True:
                        self.pc.create_index(
                            name=compliant_index_name,
                            dimension=index_meta[0]["dimensions"],
                            metric=standardize_metric_reverse(
                                index_meta[0]["metric"], "pinecone"
                            ),
                            spec=ServerlessSpec(
                                cloud=self.args["cloud"],
                                region=self.args["region"],
                            ),
                        )
                    else:
                        self.pc.create_index(
                            name=compliant_index_name,
                            dimension=index_meta[0]["dimensions"],
                            metric=standardize_metric_reverse(
                                index_meta[0]["metric"], "pinecone"
                            ),
                            spec=PodSpec(
                                environment=self.args["environment"],
                                pod_type=(
                                    self.args["pod_type"]
                                    if (
                                        "pod_type" in self.args
                                        and self.args["pod_type"] is not None
                                    )
                                    else "starter"
                                ),
                            ),
                        )
                except Exception as e:
                    tqdm.write(f"{e}")
                    raise Exception(f"Invalid index name '{compliant_index_name}'", e)
            index = self.pc.Index(compliant_index_name)
            BATCH_SIZE = self.args.get("batch_size", 1000) or 1000
            current_batch_size = BATCH_SIZE
            for namespace_meta in tqdm(index_meta, desc="Importing namespaces"):
                tqdm.write(
                    f"Importing data for namespace '{namespace_meta['namespace']}'"
                )
                namespace = namespace_meta["namespace"]
                data_path = namespace_meta["data_path"]

                # Check if the data path exists
                final_data_path = self.get_final_data_path(data_path)

                # Load the data from the parquet files
                parquet_files = self.get_parquet_files(final_data_path)

                vectors = {}
                metadata = {}
                vector_column_names, vector_column_name = self.get_vector_column_name(
                    index_name, namespace_meta
                )

                for file in tqdm(parquet_files, desc="Loading data from parquet files"):
                    file_path = os.path.join(final_data_path, file)
                    df = self.read_parquet_progress(file_path)

                    if self.args["subset"] is True:
                        if (
                            "id_list_file" in self.args
                            and self.args["id_list_file"] is not None
                        ):
                            id_list = pd.read_csv(
                                self.args["id_list_file"], header=None
                            )[0].tolist()
                            df = df[df[self.id_column].isin(id_list)]
                        elif (
                            "id_range_start" in self.args
                            and self.args["id_range_start"] is not None
                            and "id_range_end" in self.args
                            and self.args["id_range_end"] is not None
                        ):
                            # convert id to int before comparison
                            df = df[
                                (
                                    df[self.id_column].astype(int)
                                    >= self.args["id_range_start"]
                                )
                                & (
                                    df[self.id_column].astype(int)
                                    <= self.args["id_range_end"]
                                )
                            ]
                        else:
                            raise Exception(
                                "Invalid arguments for subset export. "
                                "Please provide either id_list_file or id_range_start and id_range_end"
                            )
                    if len(vectors) > (self.args.get("max_num_rows") or INT_MAX):
                        max_hit = True
                        break
                    if len(vectors) + len(df) > (
                        self.args.get("max_num_rows") or INT_MAX
                    ):
                        df = df.head(
                            (self.args.get("max_num_rows") or INT_MAX) - len(vectors)
                        )
                        max_hit = True
                    self.update_vectors(vectors, vector_column_name, df)
                    self.update_metadata(metadata, vector_column_names, df)
                    if max_hit:
                        break
                tqdm.write(
                    f"Loaded {len(vectors)} vectors from {len(parquet_files)} parquet files"
                )
                # Upsert the vectors and metadata to the Pinecone index in batches
                self.total_imported_count = 0
                start_idx = 0
                pbar = tqdm(total=len(vectors), desc="Upserting vectors")
                while start_idx < len(vectors):
                    end_idx = min(start_idx + current_batch_size, len(vectors))

                    batch_vectors = [
                        (
                            Vector(
                                id=str(id),
                                values=vector,
                                metadata={
                                    k: v
                                    for k, v in metadata.get(id, {}).items()
                                    if v is not None
                                },
                            )
                            if len(metadata.get(id, {}).keys()) > 0
                            else Vector(
                                id=str(id),
                                values=vector,
                            )
                        )
                        for id, vector in list(vectors.items())[start_idx:end_idx]
                    ]
                    try:
                        resp = index.upsert(vectors=batch_vectors, namespace=namespace)
                        self.total_imported_count += resp["upserted_count"]
                        pbar.update(resp["upserted_count"])
                        start_idx += resp["upserted_count"]
                    except Exception as e:
                        tqdm.write(
                            f"Error upserting vectors for index '{compliant_index_name}', {e}"
                        )
                        if current_batch_size < BATCH_SIZE / 100:
                            tqdm.write("Batch size is not the issue. Aborting import")
                            raise e
                        current_batch_size = int(2 * current_batch_size / 3)
                        tqdm.write(f"Reducing batch size to {current_batch_size}")
                        continue
        tqdm.write(
            f"Data import completed successfully. Imported {self.total_imported_count} vectors"
        )
        self.args["imported_count"] = self.total_imported_count



import ast
import datetime
from functools import lru_cache
import json
import os
import numpy as np
from packaging.version import Version
import abc
from tqdm import tqdm
from halo import Halo

from qdrant_client.http.models import Distance

import vdf_io
from vdf_io.constants import ID_COLUMN
from vdf_io.meta_types import NamespaceMeta, VDFMeta
from vdf_io.util import (
    expand_shorthand_path,
    extract_data_hash,
    get_final_data_path,
    get_parquet_files,
    read_parquet_progress,
)


class ImportVDB(abc.ABC):
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if not hasattr(cls, "DB_NAME_SLUG"):
            raise TypeError(
                f"Class {cls.__name__} lacks required class variable 'DB_NAME_SLUG'"
            )

    def __init__(self, args):
        self.args = args
        self.args["hash_value"] = extract_data_hash(args)
        self.hash_value = extract_data_hash(args)
        self.temp_file_paths = []
        self.abnormal_vector_format = False
        if self.args.get("hf_dataset", None) is None:
            self.args["dir"] = expand_shorthand_path(self.args["dir"])
            if not os.path.isdir(self.args["dir"]):
                raise Exception("Invalid dir path")
            if not os.path.isfile(os.path.join(self.args["dir"], "VDF_META.json")):
                raise Exception("Invalid dir path, VDF_META.json not found")
            # Check if the VDF_META.json file exists
            vdf_meta_path = os.path.join(self.args["dir"], "VDF_META.json")
            if not os.path.isfile(vdf_meta_path):
                raise Exception("VDF_META.json not found in the specified directory")
            with open(vdf_meta_path) as f:
                self.vdf_meta = json.load(f)
        else:
            hf_dataset = self.args.get("hf_dataset", None)
            index_name = hf_dataset.split("/")[-1]
            from huggingface_hub import HfFileSystem

            fs = HfFileSystem()
            with Halo(text="Checking for VDF_META.json", spinner="dots"):
                hf_files = fs.ls(f"datasets/{hf_dataset}", detail=False)
            if f"datasets/{hf_dataset}/VDF_META.json" in hf_files:
                print(f"Found VDF_META.json in {hf_dataset} on HuggingFace Hub")
                self.vdf_meta = json.loads(
                    fs.read_text(f"datasets/{hf_dataset}/VDF_META.json")
                )
            else:
                self.vdf_meta = VDFMeta(
                    version=vdf_io.__version__,
                    file_structure=[],
                    author=hf_dataset.split("/")[0],
                    exported_from="hf",
                    exported_at=datetime.datetime.now().astimezone().isoformat(),
                    id_column=self.args.get("id_column", ID_COLUMN) or ID_COLUMN,
                    indexes={
                        index_name: [
                            NamespaceMeta(
                                namespace="",
                                index_name=index_name,
                                total_vector_count=self.args.get("max_num_rows"),
                                exported_vector_count=self.args.get("max_num_rows"),
                                dimensions=self.args.get("vector_dim", -1),
                                model_name=self.args.get("model_name", ""),
                                vector_columns=self.args.get(
                                    "vector_columns", "vector"
                                ).split(","),
                                data_path=".",
                                metric=self.args.get("metric", Distance.COSINE),
                            )
                        ]
                    },
                ).model_dump()
            print(json.dumps(self.vdf_meta, indent=4))
        self.id_column = self.vdf_meta.get("id_column", ID_COLUMN) or ID_COLUMN
        if "indexes" not in self.vdf_meta:
            raise Exception("Invalid VDF_META.json, 'indexes' key not found")
        if "version" not in self.vdf_meta:
            print("Warning: 'version' key not found in VDF_META.json")
        elif "library_version" not in self.args:
            print(
                "Warning: 'library_version' not found in args. Skipping version check."
            )
        elif Version(self.vdf_meta["version"]) > Version(self.args["library_version"]):
            print(
                f"Warning: The version of vector-io library: ({self.args['library_version']}) is behind the version of the vdf directory: ({self.vdf_meta['version']})."
            )
            print(
                "Please upgrade the vector-io library to the latest version to ensure compatibility."
            )
        print("ImportVDB initialized successfully.")

    @abc.abstractmethod
    def upsert_data():
        """
        Get data from vector database
        """
        raise NotImplementedError

    def get_vector_column_name(
        self, index_name, namespace_meta, multi_vector_supported=False
    ):
        if "vector_columns" not in namespace_meta:
            print(
                "vector_columns not found in namespace metadata. Using 'vector' as the vector column name."
            )
            vector_column_name = "vector"
            vector_column_names = [vector_column_name]
        else:
            vector_column_names = namespace_meta["vector_columns"]
            vector_column_name = vector_column_names[0]
            if len(vector_column_names) > 1 and not multi_vector_supported:
                tqdm.write(
                    f"Warning: More than one vector column found for index {index_name}."
                    f" Only the first vector column {vector_column_name} will be imported."
                )
        return vector_column_names, vector_column_name

    @lru_cache(maxsize=1)
    def get_parquet_files(self, data_path):
        return get_parquet_files(
            data_path, self.args, self.temp_file_paths, self.id_column
        )

    def get_final_data_path(self, data_path):
        return get_final_data_path(
            self.args["cwd"], self.args["dir"], data_path, self.args
        )

    def get_file_path(self, final_data_path, parquet_file):
        if self.args.get("hf_dataset", None):
            return parquet_file
        return os.path.join(final_data_path, parquet_file)

    def set_dims(self, namespace_meta, index_name):
        if namespace_meta["dimensions"] == -1:
            tqdm.write(f"Resolving dimensions for index '{index_name}'")
            dims = self.resolve_dims(namespace_meta, index_name)
            if dims != -1:
                namespace_meta["dimensions"] = dims
                tqdm.write(f"Resolved dimensions: {dims}")
            else:
                tqdm.write(f"Failed to resolve dimensions for index '{index_name}'")
                raise ValueError(
                    f"Failed to resolve dimensions for index '{index_name}'"
                )

    def resolve_dims(self, namespace_meta, index_name):
        final_data_path = self.get_final_data_path(namespace_meta["data_path"])
        parquet_files = self.get_parquet_files(final_data_path)
        _, vector_column_name = self.get_vector_column_name(index_name, namespace_meta)
        dims = -1
        for file in tqdm(parquet_files, desc="Iterating parquet files"):
            file_path = self.get_file_path(final_data_path, file)
            df = self.read_parquet_progress(file_path, columns=[vector_column_name])
            i = 0
            while i < len(df[vector_column_name]):
                first_el = df[vector_column_name].iloc[i]
                if first_el is None:
                    i += 1
                    continue
                dims = len(self.extract_vector(first_el))
                break
        return dims

    def extract_vector(self, v):
        ret_v = None
        if isinstance(v, list) and len(v) > 1:
            ret_v = v
        elif isinstance(v, np.ndarray):
            if v.shape[0] > 1:
                self.abnormal_vector_format = True
                ret_v = v.tolist()
            if v.shape[0] == 1:
                self.abnormal_vector_format = True
                ret_v = v[0].tolist()
        elif isinstance(v, bytes):
            self.abnormal_vector_format = True
            tqdm.write("Warning: Vector is in bytes format. Converting to list")
            ret_v = ast.literal_eval(v.decode("utf-8"))
        elif isinstance(v, str):
            self.abnormal_vector_format = True
            ret_v = ast.literal_eval(v)
        else:
            ret_v = v
        # convert each element to float
        if ret_v is not None:
            ret_v = [float(x) for x in ret_v]
        return ret_v

    def update_metadata(self, metadata, vector_column_names, df):
        metadata.update(
            {
                row[self.id_column]: {
                    key: value
                    for key, value in row.items()
                    if key not in vector_column_names
                }
                for _, row in df.iterrows()
                if self.id_column in row
            }
        )

    def update_vectors(self, vectors, vector_column_name, df):
        for _, row in tqdm(df.iterrows(), desc="Extracting vectors", total=len(df)):
            if self.id_column in row:
                vectors[row[self.id_column]] = self.extract_vector(
                    row[vector_column_name]
                )

    def read_parquet_progress(self, file_path, **kwargs):
        return read_parquet_progress(file_path, self.id_column, **kwargs)

    def create_new_name(self, index_name, indexes, delimiter="-"):
        if not self.args.get("create_new", False):
            return index_name

        # Original name to use as the base for appending suffixes
        og_name = index_name

        # Find all indexes that start with the original name followed by a hyphen
        suffixes = [
            name[len(og_name) + 1 :]
            for name in indexes
            if name.startswith(og_name + delimiter)
        ]

        # Convert suffixes to integers where possible
        suffixes = [int(suffix) for suffix in suffixes if suffix.isdigit()]

        # Determine the next suffix to use
        suffix = max(suffixes) + 1 if suffixes else 2

        # Generate new names until a unique one is found
        while True:
            new_name = og_name + f"{delimiter}{suffix}"
            if new_name not in indexes:
                return new_name
            suffix += 1

    # destructor
    def cleanup(self):
        for temp_file_path in self.temp_file_paths:
            if os.path.isfile(temp_file_path):
                os.remove(temp_file_path)



from typing import Dict, List
from dotenv import load_dotenv
import pandas as pd
from tqdm import tqdm
import pyarrow.parquet as pq

import lancedb

from vdf_io.constants import INT_MAX
from vdf_io.meta_types import NamespaceMeta
from vdf_io.names import DBNames
from vdf_io.util import (
    cleanup_df,
    set_arg_from_input,
    set_arg_from_password,
)
from vdf_io.import_vdf.vdf_import_cls import ImportVDB


load_dotenv()


class ImportLanceDB(ImportVDB):
    DB_NAME_SLUG = DBNames.LANCEDB

    @classmethod
    def import_vdb(cls, args):
        """
        Import data to LanceDB
        """
        set_arg_from_input(
            args,
            "endpoint",
            "Enter the URL of LanceDB instance (default: '~/.lancedb'): ",
            str,
            default_value="~/.lancedb",
        )
        set_arg_from_password(
            args,
            "lancedb_api_key",
            "Enter the LanceDB API key (default: value of os.environ['LANCEDB_API_KEY']): ",
            "LANCEDB_API_KEY",
        )
        lancedb_import = ImportLanceDB(args)
        lancedb_import.upsert_data()
        return lancedb_import

    @classmethod
    def make_parser(cls, subparsers):
        parser_lancedb = subparsers.add_parser(
            cls.DB_NAME_SLUG, help="Import data to lancedb"
        )
        parser_lancedb.add_argument(
            "--endpoint", type=str, help="Location of LanceDB instance"
        )
        parser_lancedb.add_argument(
            "--lancedb_api_key", type=str, help="LanceDB API key"
        )
        parser_lancedb.add_argument(
            "--tables", type=str, help="LanceDB tables to export (comma-separated)"
        )

    def __init__(self, args):
        # call super class constructor
        super().__init__(args)
        self.db = lancedb.connect(
            self.args["endpoint"], api_key=self.args.get("lancedb_api_key") or None
        )

    def upsert_data(self):
        max_hit = False
        self.total_imported_count = 0
        indexes_content: Dict[str, List[NamespaceMeta]] = self.vdf_meta["indexes"]
        index_names: List[str] = list(indexes_content.keys())
        if len(index_names) == 0:
            raise ValueError("No indexes found in VDF_META.json")
        tables = self.db.table_names()
        # Load Parquet file
        # print(indexes_content[index_names[0]]):List[NamespaceMeta]
        for index_name, index_meta in tqdm(
            indexes_content.items(), desc="Importing indexes"
        ):
            for namespace_meta in tqdm(index_meta, desc="Importing namespaces"):
                self.set_dims(namespace_meta, index_name)
                data_path = namespace_meta["data_path"]
                final_data_path = self.get_final_data_path(data_path)
                # Load the data from the parquet files
                parquet_files = self.get_parquet_files(final_data_path)

                vectors_all = {}
                for vec_col in namespace_meta.get("vector_columns", []):
                    vectors_all[vec_col] = {}

                new_index_name = index_name + (
                    f'_{namespace_meta["namespace"]}'
                    if namespace_meta["namespace"]
                    else ""
                )
                new_index_name = self.create_new_name(new_index_name, tables)
                if new_index_name not in tables:
                    table = self.db.create_table(
                        new_index_name, schema=pq.read_schema(parquet_files[0])
                    )
                else:
                    table = self.db.open_table(new_index_name)

                for file in tqdm(parquet_files, desc="Iterating parquet files"):
                    file_path = self.get_file_path(final_data_path, file)
                    df = self.read_parquet_progress(
                        file_path,
                        max_num_rows=(
                            (self.args.get("max_num_rows") or INT_MAX)
                            - self.total_imported_count
                        ),
                    )
                    df = cleanup_df(df)
                    # if there are additional columns in the parquet file, add them to the table
                    for col in df.columns:
                        if col not in [field.name for field in table.schema]:
                            col_type = df[col].dtype
                            table.add_columns(
                                {
                                    col: get_default_value(col_type),
                                }
                            )
                    # split in batches
                    BATCH_SIZE = self.args.get("batch_size") or 1000
                    for batch in tqdm(
                        divide_into_batches(df, BATCH_SIZE),
                        desc="Importing batches",
                        total=len(df) // BATCH_SIZE,
                    ):
                        if self.total_imported_count + len(batch) >= (
                            self.args.get("max_num_rows") or INT_MAX
                        ):
                            batch = batch[
                                : (self.args.get("max_num_rows") or INT_MAX)
                                - self.total_imported_count
                            ]
                            max_hit = True
                        # convert df into list of dicts
                        batch_dict = batch.to_dict(orient="records")
                        table.add(batch_dict)
                        self.total_imported_count += len(batch)
                        if max_hit:
                            break
                tqdm.write(f"Imported {self.total_imported_count} rows")
                tqdm.write(f"New table size: {table.count_rows()}")
                if max_hit:
                    break
        print("Data imported successfully")


def divide_into_batches(df, batch_size):
    """
    Divide the dataframe into batches of size batch_size
    """
    for i in range(0, len(df), batch_size):
        yield df[i : i + batch_size]


def get_default_value(data_type):
    # Define default values for common data types
    default_values = {
        "object": "",
        "int64": 0,
        "float64": 0.0,
        "bool": False,
        "datetime64[ns]": pd.Timestamp("NaT"),
        "timedelta64[ns]": pd.Timedelta("NaT"),
    }
    # Return the default value for the specified data type, or None if not specified
    return default_values.get(data_type.name, None)



from pydantic import BaseModel, ConfigDict
from typing import Any, Dict, List, Optional


class NamespaceMeta(BaseModel):
    namespace: str
    index_name: str
    total_vector_count: int
    exported_vector_count: int
    dimensions: int
    model_name: str | None = None
    model_map: Dict[str, Any] | None = None
    vector_columns: List[str] = ["vector"]
    data_path: str
    metric: str | None = None
    index_config: Optional[Dict[Any, Any]] = None
    # schema_dict is a byte string
    schema_dict_str: str | None = None
    model_config = ConfigDict(protected_namespaces=())


class VDFMeta(BaseModel):
    version: str
    file_structure: List[str]
    author: str
    exported_from: str
    indexes: Dict[str, List[NamespaceMeta]]
    exported_at: str
    id_column: Optional[str] = None
